@@ -1,16 +1,19 @@
+import copy
 import dataclasses
 import functools
 import math
 from dataclasses import InitVar
 from functools import cached_property
-from typing import Any, Callable, Optional, Iterable, Union
+from typing import Callable, Optional, Iterable, Union, Any, Final
 
 import clingo
 import clingo.ast
 import typeguard
+from dumbo_asp.utils import extract_parsed_string
+from dumbo_utils.primitives import PrivateKey
+from dumbo_utils.validation import validate, ValidationError
 
 from dumbo_asp import utils
-from dumbo_asp.utils import validate
 
 
 @typeguard.typechecked
@@ -23,11 +26,11 @@ class Parser:
         end: int
         message: str
 
-        key: InitVar[Any]
-        __key = object()
+        key: InitVar[PrivateKey]
+        __key = PrivateKey()
 
-        def __post_init__(self, key: Any):
-            validate("key", key, equals=self.__key, help_msg="Use a factory method")
+        def __post_init__(self, key: PrivateKey):
+            self.__key.validate(key)
 
         @staticmethod
         def parse(error: str, parsed_string: str) -> "Parser.Error":
@@ -82,18 +85,18 @@ class Parser:
 
     @staticmethod
     def parse_program(string: str) -> list[clingo.ast.AST]:
-        res = []
         def callback(ast):
-            res.append(ast)
+            callback.res.append(ast)
+        callback.res = []
 
         messages = []
         try:
             clingo.ast.parse_string(string, callback, logger=lambda code, message: messages.append((code, message)))
-            validate("nonempty res", res, min_len=1)
-            validate("base program", res[0].ast_type == clingo.ast.ASTType.Program and res[0].name == "base" and
-                     len(res[0].parameters) == 0, equals=True)
-            validate("only rules", [x for x in res[1:] if x.ast_type != clingo.ast.ASTType.Rule], empty=True)
-            return res[1:]
+            validate("nonempty res", callback.res, min_len=1)
+            validate("base program", callback.res[0].ast_type == clingo.ast.ASTType.Program and
+                     callback.res[0].name == "base" and len(callback.res[0].parameters) == 0, equals=True)
+            validate("only rules", [x for x in callback.res[1:] if x.ast_type != clingo.ast.ASTType.Rule], empty=True)
+            return callback.res[1:]
         except RuntimeError:
             errors = [message[1] for message in messages if message[0] == clingo.MessageCode.RuntimeError]
             validate("errors", messages, length=1)
@@ -106,13 +109,13 @@ class Parser:
 class Predicate:
     name: str
     arity: Optional[int]
-    key: InitVar[Any]
+    key: InitVar[PrivateKey]
 
-    __key = object()
+    __key = PrivateKey()
     MAX_ARITY = 999
 
-    def __post_init__(self, key: Any):
-        validate("key", key, equals=self.__key, help_msg="Use a factory method")
+    def __post_init__(self, key: PrivateKey):
+        self.__key.validate(key)
 
     @staticmethod
     def parse(name: str, arity: Optional[int] = None) -> "Predicate":
@@ -225,16 +228,73 @@ class GroundAtom:
 
 @typeguard.typechecked
 @dataclasses.dataclass(frozen=True)
-class SymbolicAtom:
-    value: clingo.ast.AST
+class SymbolicTerm:
+    __value: clingo.ast.AST
+    __parsed_string: Optional[str]
 
-    def __post_init__(self):
-        validate("type", self.value.ast_type, equals=clingo.ast.ASTType.SymbolicAtom)
+    key: InitVar[PrivateKey]
+    __key = PrivateKey()
+
+    def __post_init__(self, key: PrivateKey):
+        self.__key.validate(key)
+        validate("type", self.__value.ast_type, equals=clingo.ast.ASTType.SymbolicTerm)
+
+    @staticmethod
+    def parse(string: str) -> "SymbolicTerm":
+        rule: Final = f":- a({string})."
+        try:
+            program = Parser.parse_program(rule)
+        except Parser.Error as error:
+            raise error.drop(first=3, last=1)
+
+        validate("one rule", program, length=1,
+                 help_msg=f"Unexpected sequence of {len(program)} rules in {utils.one_line(string)}")
+        validate("one atom", program[0].body, length=1,
+                 help_msg=f"Unexpected conjunction of {len(program[0].body)} atoms in {utils.one_line(string)}")
+        atom = program[0].body[0].atom.symbol
+        validate("arity", atom.arguments, length=1,
+                 help_msg=f"Unexpected sequence of {len(atom.arguments)} terms in {utils.one_line(string)}")
+        return SymbolicTerm(atom.arguments[0], utils.extract_parsed_string(rule, atom.arguments[0].location),
+                            key=SymbolicTerm.__key)
+
+    @staticmethod
+    def of_int(value: int) -> "SymbolicTerm":
+        return SymbolicTerm.parse(str(value))
+
+    @staticmethod
+    def of_string(value: str) -> "SymbolicTerm":
+        return SymbolicTerm.parse(f'"{value}"')
+
+    def __str__(self):
+        return self.__parsed_string or str(self.__value)
+
+    def make_copy_of_value(self) -> clingo.ast.AST:
+        return copy.deepcopy(self.__value)
+
+
+@typeguard.typechecked
+@dataclasses.dataclass(frozen=True)
+class SymbolicAtom:
+    __value: clingo.ast.AST
+    __parsed_string: Optional[str]
+
+    key: InitVar[PrivateKey]
+    __key = PrivateKey()
+
+    def __post_init__(self, key: PrivateKey):
+        self.__key.validate(key)
+        validate("type", self.__value.ast_type,
+                 is_in=[clingo.ast.ASTType.SymbolicAtom, clingo.ast.ASTType.BooleanConstant])
+
+    @staticmethod
+    def of_false() -> "SymbolicAtom":
+        return SymbolicAtom.parse("#false")
 
     @staticmethod
     def parse(string: str) -> "SymbolicAtom":
+        rule: Final = f":- {string}."
         try:
-            program = Parser.parse_program(f":- {string}.")
+            program = Parser.parse_program(rule)
         except Parser.Error as error:
             raise error.drop(first=3, last=1)
 
@@ -246,39 +306,164 @@ class SymbolicAtom:
         validate("positive", literal.sign, equals=clingo.ast.Sign.NoSign,
                  help_msg=f"Unexpected default negation in {utils.one_line(string)}")
         atom = literal.atom
-        return SymbolicAtom(atom)
+        return SymbolicAtom(atom, utils.extract_parsed_string(rule, literal.location), key=SymbolicAtom.__key)
 
     def __str__(self):
-        return str(self.value)
+        return self.__parsed_string or str(self.__value)
+
+    def make_copy_of_value(self) -> clingo.ast.AST:
+        return copy.deepcopy(self.__value)
 
 
 @typeguard.typechecked
 @dataclasses.dataclass(frozen=True)
 class SymbolicRule:
-    value: clingo.ast.AST
-    parsed_string: Optional[str]
+    __value: clingo.ast.AST
+    __parsed_string: Optional[str]
 
-    def __post_init__(self):
-        validate("type", self.value.ast_type, equals=clingo.ast.ASTType.Rule)
+    key: InitVar[PrivateKey]
+    __key = PrivateKey()
+
+    def __post_init__(self, key: PrivateKey):
+        self.__key.validate(key)
+        validate("type", self.__value.ast_type, equals=clingo.ast.ASTType.Rule)
 
     @staticmethod
     def parse(string: str) -> "SymbolicRule":
         program = Parser.parse_program(string)
         validate("one rule", program, length=1,
                  help_msg=f"Unexpected sequence of {len(program)} rules in {utils.one_line(string)}")
-        return SymbolicRule(program[0], string)
+        return SymbolicRule(program[0], utils.extract_parsed_string(string, program[0].location),
+                            key=SymbolicRule.__key)
+
+    @staticmethod
+    def of(value: clingo.ast.AST) -> "SymbolicRule":
+        validate("value", value.ast_type == clingo.ast.ASTType.Rule, equals=True)
+        return SymbolicRule(value, None, key=SymbolicRule.__key)
 
     def __str__(self):
-        return str(self.value)
+        return self.__parsed_string or str(self.__value)
+
+    def transform(self, transformer: clingo.ast.Transformer) -> Any:
+        transformer(self.__value)
+
+    @cached_property
+    def head_variables(self) -> tuple[str, ...]:
+        res = set()
+
+        class Transformer(clingo.ast.Transformer):
+            def visit_Variable(self, node):
+                res.add(str(node))
+                return node
+
+        Transformer().visit(self.__value.head)
+        return tuple(sorted(res))
+
+    @cached_property
+    def body_variables(self) -> tuple[str, ...]:
+        res = set()
+
+        class Transformer(clingo.ast.Transformer):
+            def visit_Variable(self, node):
+                res.add(str(node))
+                return node
+
+        Transformer().visit_sequence(self.__value.body)
+        return tuple(sorted(res))
+
+    @cached_property
+    def global_safe_variables(self) -> tuple[str, ...]:
+        res = set()
+
+        class Transformer(clingo.ast.Transformer):
+            def visit_Literal(self, node):
+                if node.sign == clingo.ast.Sign.NoSign:
+                    self.visit_children(node)
+
+            def visit_BodyAggregate(self, node):
+                for guard in [node.left_guard, node.right_guard]:
+                    if guard.comparison == clingo.ast.ComparisonOperator.Equal:
+                        self.visit(guard.term)
+
+            def visit_Variable(self, node):
+                res.add(str(node))
+                return node
+
+        Transformer().visit_sequence(self.__value.body)
+        return tuple(sorted(res))
+
+    def with_extended_body(self, atom: SymbolicAtom, sign: clingo.ast.Sign = clingo.ast.Sign.NoSign) -> "SymbolicRule":
+        string = self.__parsed_string[:-1] if self.__parsed_string is not None else str(self)[:-1]
+        literal = f"{atom}" if sign == clingo.ast.Sign.NoSign else \
+            f"not {atom}" if sign == clingo.ast.Sign.Negation else \
+            f"not not {atom}"
+        return self.parse(f"{string}; {literal}." if len(self.__value.body) > 0 else f"{string} :- {literal}.")
+
+    def body_as_string(self, separator: str = "; ") -> str:
+        return separator.join(str(x) for x in self.__value.body)
+
+    def apply_variable_substitution(self, **kwargs: SymbolicTerm) -> "SymbolicRule":
+        class Transformer(clingo.ast.Transformer):
+            def visit_Variable(self, node):
+                if str(node) not in kwargs.keys():
+                    return node
+                return kwargs[str(node)].make_copy_of_value()
+
+        return self.of(Transformer().visit(self.__value))
+
+
+@typeguard.typechecked
+@dataclasses.dataclass(frozen=True)
+class SymbolicProgram:
+    __rules: tuple[SymbolicRule, ...]
+    __parsed_string: Optional[str]
+
+    key: InitVar[PrivateKey]
+    __key = PrivateKey()
+
+    def __post_init__(self, key: PrivateKey):
+        self.__key.validate(key)
+
+    @staticmethod
+    def of(*args: SymbolicRule | Iterable[SymbolicRule]) -> "SymbolicProgram":
+        rules = []
+        for arg in args:
+            if type(arg) == SymbolicRule:
+                rules.append(arg)
+            else:
+                rules.extend(arg)
+        return SymbolicProgram(tuple(rules), None, key=SymbolicProgram.__key)
+
+    @staticmethod
+    def parse(string: str) -> "SymbolicProgram":
+        rules = tuple(SymbolicRule.parse(utils.extract_parsed_string(string, rule.location))
+                      for rule in Parser.parse_program(string))
+        return SymbolicProgram(rules, string, key=SymbolicProgram.__key)
+
+    def __str__(self):
+        return '\n'.join(str(rule) for rule in self.__rules) if self.__parsed_string is None else self.__parsed_string
+
+    def __len__(self):
+        return len(self.__rules)
+
+    def __getitem__(self, item: int):
+        return self.__rules[item]
+
+    @cached_property
+    def herbrand_base(self) -> "Model":
+        control = clingo.Control()
+        control.add(str(self))
+        control.ground([("base", [])])
+        return Model.of_atoms(atom.symbol for atom in control.symbolic_atoms)
 
 
 @typeguard.typechecked
 @dataclasses.dataclass(frozen=True, order=True)
 class Model:
-    key: dataclasses.InitVar[Any]
-    value: tuple[GroundAtom, ...]
+    value: tuple[GroundAtom | int | str, ...]
 
-    __key = object()
+    key: InitVar[PrivateKey]
+    __key = PrivateKey()
 
     class NoModelError(ValueError):
         def __init__(self, *args):
@@ -293,16 +478,12 @@ class Model:
         return Model(key=Model.__key, value=())
 
     @staticmethod
-    def of(control: clingo.Control) -> "Model":
+    def of_control(control: clingo.Control) -> "Model":
         def on_model(model):
             if on_model.cost is not None and on_model.cost <= model.cost:
                 on_model.exception = True
             on_model.cost = model.cost
-            atoms = [GroundAtom(x) for x in model.symbols(shown=True)]
-            on_model.res = Model(
-                key=Model.__key,
-                value=tuple(sorted(atoms)),
-            )
+            on_model.res = Model.of_elements(model.symbols(shown=True))
         on_model.cost = None
         on_model.res = None
         on_model.exception = False
@@ -315,22 +496,47 @@ class Model:
         return on_model.res
 
     @staticmethod
-    def of_program(program: str) -> "Model":
+    def of_program(*args: str | SymbolicProgram | Iterable[str | SymbolicProgram]) -> "Model":
+        program = []
+        for arg in args:
+            if type(arg) is str:
+                program.append(arg)
+            elif type(arg) is SymbolicProgram:
+                program.append(str(arg))
+            else:
+                program.extend(str(elem) for elem in arg)
         control = clingo.Control()
-        control.add("base", [], program)
+        control.add('\n'.join(program))
         control.ground([("base", [])])
-        return Model.of(control)
+        return Model.of_control(control)
 
     @staticmethod
     def of_atoms(*args: Union[str, clingo.Symbol, GroundAtom,
                               Iterable[Union[str, clingo.Symbol, GroundAtom]]]) -> "Model":
+        res = Model.of_elements(*args)
+        validate("only atoms", res.contains_only_ground_atoms, equals=True,
+                 help_msg="Use Model.of_elements() to create a model with numbers and strings")
+        return res
+
+    @staticmethod
+    def of_elements(*args: Union[int, str, clingo.Symbol, GroundAtom,
+                    Iterable[Union[int, str, clingo.Symbol, GroundAtom]]]) -> "Model":
         def build(atom):
-            if type(atom) is GroundAtom:
+            if type(atom) in [GroundAtom, int]:
                 return atom
             if type(atom) is clingo.Symbol:
+                if atom.type == clingo.SymbolType.Number:
+                    return atom.number
+                if atom.type == clingo.SymbolType.String:
+                    return atom.string
                 return GroundAtom(atom)
             if type(atom) is str:
-                return GroundAtom.parse(atom)
+                try:
+                    return GroundAtom.parse(atom)
+                except ValidationError:
+                    if atom[0] == '"' == atom[-1]:
+                        return Parser.parse_ground_term(atom).string
+                    return Parser.parse_ground_term(f'"{atom}"').string
             return None
 
         flattened = []
@@ -340,10 +546,16 @@ class Model:
                 flattened.append(built_element)
             else:
                 flattened.extend(build(atom) for atom in element)
-        return Model(key=Model.__key, value=tuple(flattened))
+        return Model(
+            key=Model.__key,
+            value=
+            tuple(sorted(x for x in flattened if type(x) is int)) +
+            tuple(sorted(x for x in flattened if type(x) is str)) +
+            tuple(sorted(x for x in flattened if type(x) is GroundAtom))
+        )
 
-    def __post_init__(self, key: Any):
-        validate("create key", key, equals=self.__key, help_msg="Use a factory method.")
+    def __post_init__(self, key: PrivateKey):
+        self.__key.validate(key)
 
     def __str__(self):
         return ' '.join(str(x) for x in self.value)
@@ -357,12 +569,30 @@ class Model:
     def __iter__(self):
         return self.value.__iter__()
 
+    @cached_property
+    def contains_only_ground_atoms(self) -> bool:
+        return all(type(element) == GroundAtom for element in self)
+
     @property
     def as_facts(self) -> str:
-        return '\n'.join(f"{atom}." for atom in self)
+        def build(element):
+            if type(element) is int:
+                return f"__number({element})."
+            if type(element) is str:
+                return f"__string(\"{element}\")."
+            return f"{element}."
+        return '\n'.join(build(element) for element in self)
 
-    def drop(self, predicate: Predicate) -> "Model":
-        return Model(key=self.__key, value=tuple(atom for atom in self if not predicate.match(atom.predicate)))
+    def drop(self, predicate: Optional[Predicate] = None, numbers: bool = False, strings: bool = False) -> "Model":
+        def when(element):
+            if type(element) is GroundAtom:
+                return predicate is None or not predicate.match(element.predicate)
+            if type(element) is int:
+                return not numbers
+            assert type(element) is str
+            return not strings
+
+        return self.filter(when)
 
     def filter(self, when: Callable[[GroundAtom], bool]) -> "Model":
         return Model(key=self.__key, value=tuple(atom for atom in self if when(atom)))
@@ -373,8 +603,9 @@ class Model:
     def rename(self, predicate: Predicate, new_name: Predicate) -> "Model":
         validate("same arity", predicate.arity == new_name.arity, equals=True,
                  help_msg="Predicates must have the same arity")
-        return self.map(lambda atom: atom if not predicate.match(atom.predicate)
-            else GroundAtom(clingo.Function(new_name.name, atom.arguments)))
+        return self.map(lambda atom: atom if not predicate.match(atom.predicate) else GroundAtom(
+            clingo.Function(new_name.name, atom.arguments)
+        ))
 
     def substitute(self, predicate: Predicate, argument: int, term: clingo.Symbol) -> "Model":
         validate("argument", argument, min_value=1, max_value=predicate.arity, help_msg="Arguments are indexed from 1")
@@ -383,7 +614,7 @@ class Model:
             if not predicate.match(atom.predicate):
                 return atom
             return GroundAtom(clingo.Function(
-                atom.name,
+                atom.predicate_name,
                 [arg if index != argument else term for index, arg in enumerate(atom.arguments, start=1)]
             ))
         return self.map(mapping)
@@ -395,7 +626,7 @@ class Model:
             if not predicate.match(atom.predicate):
                 return atom
             return GroundAtom(clingo.Function(
-                atom.name,
+                atom.predicate_name,
                 [arg for index, arg in enumerate(atom.arguments, start=1) if index != argument]
             ))
         return self.map(mapping)
